# Copyright (c) 2012 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Request Body limiting middleware.
"""

from __future__ import annotations

import logging
import typing as ty

from oslo_config import cfg
import webob.dec
import webob.exc

from oslo_middleware._i18n import _
from oslo_middleware import base

if ty.TYPE_CHECKING:
    from _typeshed.wsgi import WSGIApplication

LOG = logging.getLogger(__name__)

OPTS = [
    # default request size is 112k
    cfg.IntOpt(
        'max_request_body_size',
        default=114688,
        help='The maximum body size for each request, in bytes.',
    ),
]


class LimitingReader:
    """Reader to limit the size of an incoming request."""

    def __init__(self, data: ty.IO[bytes], limit: int) -> None:
        """Initiates LimitingReader object.

        :param data: Underlying data object
        :param limit: maximum number of bytes the reader should allow
        """
        self.data = data
        self.limit = limit
        self.bytes_read = 0

    def __iter__(self) -> ty.Iterator[bytes]:
        for chunk in self.data:
            self.bytes_read += len(chunk)
            if self.bytes_read > self.limit:
                msg = _("Request is too large. Larger than %s") % self.limit
                raise webob.exc.HTTPRequestEntityTooLarge(explanation=msg)
            else:
                yield chunk

    def read(self, i: int | None = None) -> bytes:
        # NOTE(jamielennox): We can't simply provide the default to the read()
        # call as the expected default differs between mod_wsgi and eventlet
        if i is None:
            result = self.data.read()
        else:
            result = self.data.read(i)
        self.bytes_read += len(result)
        if self.bytes_read > self.limit:
            msg = _("Request is too large. Larger than %s.") % self.limit
            raise webob.exc.HTTPRequestEntityTooLarge(explanation=msg)
        return result


class RequestBodySizeLimiter(base.ConfigurableMiddleware):
    """Limit the size of incoming requests."""

    def __init__(
        self,
        application: WSGIApplication | None,
        conf: dict[str, ty.Any] | cfg.ConfigOpts | None = None,
    ) -> None:
        super().__init__(application, conf)
        self.oslo_conf.register_opts(OPTS, group='oslo_middleware')

    @webob.dec.wsgify
    def __call__(
        self,
        req: webob.request.Request,
    ) -> webob.response.Response | None:
        max_size = self._conf_get('max_request_body_size')
        if req.content_length is not None and req.content_length > max_size:
            msg = _(
                "Request is too large. Larger than max_request_body_size (%s)."
            )
            LOG.info(msg, max_size)
            raise webob.exc.HTTPRequestEntityTooLarge(explanation=msg)

        if req.content_length is None:
            # the type stub for this is incomplete: body_file is not just
            # readable - it's also iterable
            limiter = LimitingReader(req.body_file, max_size)  # type: ignore
            req.body_file = limiter

        return req.get_response(self.application)



# Copyright (c) 2015 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from wsgiref import util

from oslotest import base as test_base
import webob

from oslo_middleware import http_proxy_to_wsgi


class TestHTTPProxyToWSGI(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()

        @webob.dec.wsgify()
        def fake_app(req):
            return util.application_uri(req.environ)

        self.middleware = http_proxy_to_wsgi.HTTPProxyToWSGI(fake_app)
        self.middleware.oslo_conf.set_override(
            'enable_proxy_headers_parsing', True, group='oslo_middleware'
        )
        self.request = webob.Request.blank('/foo/bar', method='POST')

    def test_backward_compat(self):
        @webob.dec.wsgify()
        def fake_app(req):
            return util.application_uri(req.environ)

        self.middleware = http_proxy_to_wsgi.HTTPProxyToWSGI(fake_app)
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"http://localhost:80/", response.body)

    def test_no_headers(self):
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"http://localhost:80/", response.body)

    def test_url_translate_ssl(self):
        self.request.headers['X-Forwarded-Proto'] = "https"
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"https://localhost:80/", response.body)

    def test_url_translate_ssl_port(self):
        self.request.headers['X-Forwarded-Proto'] = "https"
        self.request.headers['X-Forwarded-Host'] = "example.com:123"
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"https://example.com:123/", response.body)

    def test_url_translate_host_ipv6(self):
        self.request.headers['X-Forwarded-Proto'] = "https"
        self.request.headers['X-Forwarded-Host'] = "[f00:b4d::1]:123"
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"https://[f00:b4d::1]:123/", response.body)

    def test_url_translate_base(self):
        self.request.headers['X-Forwarded-Prefix'] = "/bla"
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"http://localhost:80/bla", response.body)

    def test_url_translate_port_and_base_and_proto_and_host(self):
        self.request.headers['X-Forwarded-Proto'] = "https"
        self.request.headers['X-Forwarded-Prefix'] = "/bla"
        self.request.headers['X-Forwarded-Host'] = "example.com:8043"
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"https://example.com:8043/bla", response.body)

    def test_rfc7239_invalid(self):
        self.request.headers['Forwarded'] = (
            "iam=anattacker;metoo, I will crash you!!P;m,xx"
        )
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"http://localhost:80/", response.body)

    def test_rfc7239_proto(self):
        self.request.headers['Forwarded'] = (
            "for=foobar;proto=https, for=foobaz;proto=http"
        )
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"https://localhost:80/", response.body)

    def test__parse_rfc7239_header(self):
        expected_result = [
            {'for': 'foobar', 'proto': 'https'},
            {'for': 'foobaz', 'proto': 'http'},
        ]

        result = self.middleware._parse_rfc7239_header(
            "for=foobar;proto=https, for=foobaz;proto=http"
        )
        self.assertEqual(expected_result, result)

        result = self.middleware._parse_rfc7239_header(
            "for=foobar; proto=https, for=foobaz; proto=http"
        )
        self.assertEqual(expected_result, result)

    def test_rfc7239_proto_host(self):
        self.request.headers['Forwarded'] = (
            "for=foobar;proto=https;host=example.com, for=foobaz;proto=http"
        )
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"https://example.com/", response.body)

    def test_rfc7239_proto_host_base(self):
        self.request.headers['Forwarded'] = (
            "for=foobar;proto=https;host=example.com:8043, for=foobaz"
        )
        self.request.headers['X-Forwarded-Prefix'] = "/bla"
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"https://example.com:8043/bla", response.body)

    def test_forwarded_for_headers(self):
        @webob.dec.wsgify()
        def fake_app(req):
            return req.environ['REMOTE_ADDR']

        self.middleware = http_proxy_to_wsgi.HTTPProxyToWSGI(fake_app)
        forwarded_for_addr = '1.2.3.4'
        forwarded_addr = '8.8.8.8'

        # If both X-Forwarded-For and Fowarded headers are present, it should
        # use the Forwarded header and ignore the X-Forwarded-For header.
        self.request.headers['Forwarded'] = (
            f"for={forwarded_addr};proto=https;host=example.com:8043"
        )
        self.request.headers['X-Forwarded-For'] = forwarded_for_addr
        response = self.request.get_response(self.middleware)
        self.assertEqual(forwarded_addr.encode(), response.body)

        # Now if only X-Forwarded-For header is present, it should be used.
        del self.request.headers['Forwarded']
        response = self.request.get_response(self.middleware)
        self.assertEqual(forwarded_for_addr.encode(), response.body)


class TestHTTPProxyToWSGIDisabled(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()

        @webob.dec.wsgify()
        def fake_app(req):
            return util.application_uri(req.environ)

        self.middleware = http_proxy_to_wsgi.HTTPProxyToWSGI(fake_app)
        self.middleware.oslo_conf.set_override(
            'enable_proxy_headers_parsing', False, group='oslo_middleware'
        )
        self.request = webob.Request.blank('/foo/bar', method='POST')

    def test_no_headers(self):
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"http://localhost:80/", response.body)

    def test_url_translate_ssl_has_no_effect(self):
        self.request.headers['X-Forwarded-Proto'] = "https"
        self.request.headers['X-Forwarded-Host'] = "example.com:123"
        response = self.request.get_response(self.middleware)
        self.assertEqual(b"http://localhost:80/", response.body)



# Copyright 2024 Red Hat.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os
import typing as ty

from oslo_middleware.healthcheck import opts
from oslo_middleware.healthcheck import pluginbase

LOG = logging.getLogger(__name__)

if ty.TYPE_CHECKING:
    from oslo_config import cfg


class EnableByFilesHealthcheck(pluginbase.HealthcheckBaseExtension):
    """EnableByFilesHealthcheck healthcheck middleware plugin

    This plugin checks presence of a file at a specified location.

    Example of middleware configuration:

    .. code-block:: ini

      [app:healthcheck]
      paste.app_factory = oslo_middleware:Healthcheck.app_factory
      path = /healthcheck
      backends = enable_by_files
      enable_by_file_paths = /var/lib/glance/images/.marker,
          /var/lib/glance/os_glance_staging_store/.marker
      # set to True to enable detailed output, False is the default
      detailed = False
    """

    def __init__(
        self,
        oslo_conf: 'cfg.ConfigOpts',
        conf: dict[str, ty.Any],
    ) -> None:
        super().__init__(oslo_conf=oslo_conf, conf=conf)
        self.oslo_conf.register_opts(
            opts.ENABLE_BY_FILES_OPTS, group='healthcheck'
        )
        self.file_paths = self._conf_get('enable_by_file_paths')

    def healthcheck(self, server_port: int) -> pluginbase.HealthcheckResult:
        for file_path in self.file_paths:
            if not os.path.exists(file_path):
                LOG.warning(
                    'EnableByFiles healthcheck middleware: Path %s '
                    'is not present',
                    file_path,
                )
                return pluginbase.HealthcheckResult(
                    available=False,
                    reason="FILE PATH MISSING",
                    details=f'File path {file_path} is missing',
                )
        return pluginbase.HealthcheckResult(
            available=True,
            reason="OK",
            details='Specified file paths are available',
        )



# Copyright (c) 2013 NEC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading
import time
from unittest import mock

from oslo_config import fixture as config
from oslo_serialization import jsonutils
from oslotest import base as test_base
import requests
import webob.dec
import webob.exc

from oslo_middleware.exceptions import ConfigInvalid
from oslo_middleware import healthcheck
from oslo_middleware.healthcheck import __main__


class HealthcheckMainTests(test_base.BaseTestCase):
    def test_startup_response(self):
        server = __main__.create_server(0)
        th = threading.Thread(target=server.serve_forever)
        th.start()
        self.addCleanup(server.shutdown)
        while True:
            try:
                # Connecting on 0.0.0.0 is not allowed on windows
                # The operating system will return WSAEADDRNOTAVAIL which
                # in turn will throw a requests.ConnectionError
                r = requests.get(
                    f"http://127.0.0.1:{server.server_address[1]}",
                    timeout=10,
                )
            except requests.ConnectionError:
                # Server hasn't started up yet, try again in a few.
                time.sleep(1)
            else:
                self.assertEqual(200, r.status_code)
                break


class HealthcheckTests(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(config.Config())

    @staticmethod
    @webob.dec.wsgify
    def application(req):
        return 'Hello, World!!!'

    def _do_test_request(
        self,
        conf={},
        path='/healthcheck',
        accept='text/plain',
        method='GET',
        server_port=80,
        headers=None,
        remote_addr='127.0.0.1',
    ):
        self.app = healthcheck.Healthcheck(self.application, conf)
        req = webob.Request.blank(path, accept=accept, method=method)
        req.server_port = server_port
        if headers:
            req.headers = headers
        req.remote_addr = remote_addr
        res = req.get_response(self.app)
        return res

    def _do_test(
        self,
        conf={},
        path='/healthcheck',
        expected_code=webob.exc.HTTPOk.code,
        expected_body=b'',
        accept='text/plain',
        method='GET',
        server_port=80,
        headers=None,
        remote_addr='127.0.0.1',
    ):
        res = self._do_test_request(
            conf=conf,
            path=path,
            accept=accept,
            method=method,
            server_port=server_port,
            headers=headers,
            remote_addr=remote_addr,
        )
        self.assertEqual(expected_code, res.status_int)
        self.assertEqual(expected_body, res.body)

    def test_default(self):
        self._do_test()

    @mock.patch('oslo_middleware.healthcheck.disable_by_file.LOG')
    def test_disablefile_unconfigured(self, fake_log):
        fake_warn = fake_log.warning
        conf = {'backends': 'disable_by_file'}
        self._do_test(conf, expected_body=b'OK')
        self.assertIn('disable_by_file', self.app._backends.names())
        fake_warn.assert_called_once_with(
            'DisableByFile healthcheck middleware '
            'enabled without disable_by_file_path '
            'set'
        )

    def test_disablefile_enabled(self):
        conf = {
            'backends': 'disable_by_file',
            'disable_by_file_path': '/foobar',
        }
        self._do_test(conf, expected_body=b'OK')
        self.assertIn('disable_by_file', self.app._backends.names())

    def test_disablefile_enabled_head(self):
        conf = {
            'backends': 'disable_by_file',
            'disable_by_file_path': '/foobar',
        }
        self._do_test(
            conf,
            expected_body=b'',
            method='HEAD',
            expected_code=webob.exc.HTTPNoContent.code,
        )

    def test_disablefile_enabled_html_detailed(self):
        conf = {
            'backends': 'disable_by_file',
            'disable_by_file_path': '/foobar',
            'detailed': True,
        }
        res = self._do_test_request(conf, accept="text/html")
        self.assertIn(b'Result of 1 checks:', res.body)
        self.assertIn(b'<TD>OK</TD>', res.body)
        self.assertEqual(webob.exc.HTTPOk.code, res.status_int)

    def test_disablefile_disabled(self):
        filename = self.create_tempfiles([('test', 'foobar')])[0]
        conf = {
            'backends': 'disable_by_file',
            'disable_by_file_path': filename,
        }
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPServiceUnavailable.code,
            expected_body=b'DISABLED BY FILE',
        )
        self.assertIn('disable_by_file', self.app._backends.names())

    def test_disablefile_disabled_head(self):
        filename = self.create_tempfiles([('test', 'foobar')])[0]
        conf = {
            'backends': 'disable_by_file',
            'disable_by_file_path': filename,
        }
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPServiceUnavailable.code,
            expected_body=b'',
            method='HEAD',
        )
        self.assertIn('disable_by_file', self.app._backends.names())

    def test_disablefile_disabled_html_detailed(self):
        filename = self.create_tempfiles([('test', 'foobar')])[0]
        conf = {
            'backends': 'disable_by_file',
            'disable_by_file_path': filename,
            'detailed': True,
        }
        res = self._do_test_request(conf, accept="text/html")
        self.assertIn(b'<TD>DISABLED BY FILE</TD>', res.body)
        self.assertEqual(webob.exc.HTTPServiceUnavailable.code, res.status_int)

    def test_two_backends(self):
        filename = self.create_tempfiles([('test', 'foobar')])[0]
        conf = {
            'backends': 'disable_by_file,disable_by_file',
            'disable_by_file_path': filename,
        }
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPServiceUnavailable.code,
            expected_body=b'DISABLED BY FILE\nDISABLED BY FILE',
        )
        self.assertIn('disable_by_file', self.app._backends.names())

    def test_disable_by_port_file(self):
        filename = self.create_tempfiles([('test', 'foobar')])[0]
        conf = {
            'backends': 'disable_by_files_ports',
            'disable_by_file_paths': f"80:{filename}",
        }
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPServiceUnavailable.code,
            expected_body=b'DISABLED BY FILE',
        )
        self.assertIn('disable_by_files_ports', self.app._backends.names())

    def test_no_disable_by_port_file(self):
        filename = self.create_tempfiles([('test', 'foobar')])[0]
        conf = {
            'backends': 'disable_by_files_ports',
            'disable_by_file_paths': f"8000:{filename}",
        }
        self._do_test(
            conf, expected_code=webob.exc.HTTPOk.code, expected_body=b'OK'
        )
        self.assertIn('disable_by_files_ports', self.app._backends.names())

    def test_disable_by_port_many_files(self):
        filename = self.create_tempfiles([('test', 'foobar')])[0]
        filename2 = self.create_tempfiles([('test2', 'foobar2')])[0]
        conf = {
            'backends': 'disable_by_files_ports',
            'disable_by_file_paths': f"80:{filename},81:{filename2}",
        }
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPServiceUnavailable.code,
            expected_body=b'DISABLED BY FILE',
        )
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPServiceUnavailable.code,
            expected_body=b'DISABLED BY FILE',
            server_port=81,
        )
        self.assertIn('disable_by_files_ports', self.app._backends.names())

    def test_enablefile_disablefile_configured(self):
        conf = {'backends': 'disable_by_file,enable_by_files'}
        self.assertRaises(
            ConfigInvalid, healthcheck.Healthcheck, self.application, conf
        )

    def test_enablefile_unconfigured(self):
        conf = {'backends': 'enable_by_files'}
        self._do_test(conf, expected_body=b'OK')
        self.assertIn('enable_by_files', self.app._backends.names())

    def test_enablefile_enabled(self):
        filename = self.create_tempfiles([('.test', '.foobar')])[0]
        conf = {
            'backends': 'enable_by_files',
            'enable_by_file_paths': filename,
        }
        self._do_test(conf, expected_body=b'OK')
        self.assertIn('enable_by_files', self.app._backends.names())

    def test_enablefile_enabled_head(self):
        filename = self.create_tempfiles([('.test', '.foobar')])[0]
        conf = {
            'backends': 'enable_by_files',
            'enable_by_file_paths': filename,
        }
        self._do_test(
            conf,
            expected_body=b'',
            method='HEAD',
            expected_code=webob.exc.HTTPNoContent.code,
        )

    def test_enablefile_enabled_html_detailed(self):
        filename = self.create_tempfiles([('.test', '.foobar')])[0]
        conf = {
            'backends': 'enable_by_files',
            'enable_by_file_paths': filename,
            'detailed': True,
        }
        res = self._do_test_request(conf, accept="text/html")
        self.assertIn(b'Result of 1 checks:', res.body)
        self.assertIn(b'<TD>OK</TD>', res.body)
        self.assertEqual(webob.exc.HTTPOk.code, res.status_int)

    def test_enablefile_disabled(self):
        conf = {
            'backends': 'enable_by_files',
            'enable_by_file_paths': '.foobar',
        }
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPServiceUnavailable.code,
            expected_body=b'FILE PATH MISSING',
        )
        self.assertIn('enable_by_files', self.app._backends.names())

    def test_enablefile_disabled_head(self):
        conf = {
            'backends': 'enable_by_files',
            'enable_by_file_paths': '.foobar',
        }
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPServiceUnavailable.code,
            expected_body=b'',
            method='HEAD',
        )
        self.assertIn('enable_by_files', self.app._backends.names())

    def test_enablefile_disabled_html_detailed(self):
        conf = {
            'backends': 'enable_by_files',
            'enable_by_file_paths': '.foobar',
            'detailed': True,
        }
        res = self._do_test_request(conf, accept="text/html")
        self.assertIn(b'<TD>FILE PATH MISSING</TD>', res.body)
        self.assertEqual(webob.exc.HTTPServiceUnavailable.code, res.status_int)

    def test_json_response(self):
        expected_body = jsonutils.dumps(
            {'detailed': False, 'reasons': []}, indent=4, sort_keys=True
        ).encode('utf-8')
        self._do_test(expected_body=expected_body, accept='application/json')

    def test_source_within_allowed_ranges(self):
        conf = {'allowed_source_ranges': ['192.168.0.0/24', '192.168.1.0/24']}
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPOk.code,
            remote_addr='192.168.0.1',
        )

    def test_source_out_of_allowed_ranges(self):
        conf = {'allowed_source_ranges': ['192.168.0.0/24', '192.168.1.0/24']}
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPOk.code,
            expected_body=b'Hello, World!!!',
            remote_addr='192.168.3.1',
        )

    def test_proxied_not_ignored(self):
        self._do_test(
            {},
            expected_code=webob.exc.HTTPOk.code,
            headers={'Forwarded-For': 'http://localhost'},
        )

    def test_proxied_ignored(self):
        conf = {'ignore_proxied_requests': True}
        modern_headers = {'x-forwarded': 'https://localhost'}
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPOk.code,
            expected_body=b'Hello, World!!!',
            headers=modern_headers,
        )
        legacy_headers = {
            'x-forwarded-proto': 'https',
            'x-forwarded-host': 'localhost',
            'x-forwarded-for': '192.0.2.11',
        }
        self._do_test(
            conf,
            expected_code=webob.exc.HTTPOk.code,
            expected_body=b'Hello, World!!!',
            headers=legacy_headers,
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__all__ = [
    'BasicAuthMiddleware',
    'CatchErrors',
    'CorrelationId',
    'CORS',
    'Debug',
    'Healthcheck',
    'HTTPProxyToWSGI',
    'RequestId',
    'RequestBodySizeLimiter',
]

from oslo_middleware.basic_auth import BasicAuthMiddleware
from oslo_middleware.catch_errors import CatchErrors
from oslo_middleware.correlation_id import CorrelationId
from oslo_middleware.cors import CORS
from oslo_middleware.debug import Debug
from oslo_middleware.healthcheck import Healthcheck
from oslo_middleware.http_proxy_to_wsgi import HTTPProxyToWSGI
from oslo_middleware.request_id import RequestId
from oslo_middleware.sizelimit import RequestBodySizeLimiter



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import collections
import gc
import io
import ipaddress
import json
import platform
import socket
import sys
import traceback
import typing as ty

import jinja2
from oslo_utils import reflection
from oslo_utils import timeutils
import stevedore
import typing_extensions as ty_ext
import webob.dec
import webob.exc
import webob.response

try:
    import greenlet
except ImportError:
    greenlet = None  # type: ignore

from oslo_middleware import base
from oslo_middleware.exceptions import ConfigInvalid
from oslo_middleware.healthcheck import opts

if ty.TYPE_CHECKING:
    from _typeshed.wsgi import WSGIApplication
    from oslo_config import cfg
    from oslo_middleware.healthcheck import pluginbase


def _find_objects(t: type[ty.Any]) -> list[ty.Any]:
    return [o for o in gc.get_objects() if isinstance(o, t)]


def _expand_template(contents: str, params: dict[str, ty.Any]) -> ty.Any:
    tpl = jinja2.Template(source=contents, undefined=jinja2.StrictUndefined)
    return tpl.render(**params)


Reason = ty.TypedDict("Reason", {"reason": str, "details": str, "class": str})


class Healthcheck(base.ConfigurableMiddleware):
    """Healthcheck application used for monitoring.

    It will respond 200 with "OK" as the body. Or a 503 with the reason as the
    body if one of the backends reports an application issue.

    This is useful for the following reasons:

    * Load balancers can 'ping' this url to determine service availability.
    * Provides an endpoint that is similar to 'mod_status' in apache which
      can provide details (or no details, depending on if configured) about
      the activity of the server.
    * *(and more)*

    .. note::

        This middleware indicates that the API is accessible but it does
        indicate that it is necessarily functional or that any other API
        request will actually work.

    Example requests/responses (**not** detailed mode)::

      $ curl -i -X HEAD "http://0.0.0.0:8775/healthcheck"
      HTTP/1.1 204 No Content
      Content-Type: text/plain; charset=UTF-8
      Content-Length: 0
      Date: Fri, 11 Sep 2015 18:55:08 GMT

      $ curl -i -X GET "http://0.0.0.0:8775/healthcheck"
      HTTP/1.1 200 OK
      Content-Type: text/plain; charset=UTF-8
      Content-Length: 2
      Date: Fri, 11 Sep 2015 18:55:43 GMT

      OK

      $ curl -X GET -i -H "Accept: application/json" "http://0.0.0.0:8775/healthcheck"
      HTTP/1.0 200 OK
      Date: Wed, 24 Aug 2016 06:09:58 GMT
      Content-Type: application/json
      Content-Length: 63

      {
          "detailed": false,
          "reasons": [
              "OK"
          ]
      }

      $ curl -X GET -i -H "Accept: text/html" "http://0.0.0.0:8775/healthcheck"
      HTTP/1.0 200 OK
      Date: Wed, 24 Aug 2016 06:10:42 GMT
      Content-Type: text/html; charset=UTF-8
      Content-Length: 239

      <HTML>
      <HEAD><TITLE>Healthcheck Status</TITLE></HEAD>
      <BODY>

      <H2>Result of 1 checks:</H2>
      <TABLE bgcolor="#ffffff" border="1">
      <TBODY>
      <TR>

      <TH>
      Reason
      </TH>
      </TR>
      <TR>
          <TD>OK</TD>
      </TR>
      </TBODY>
      </TABLE>
      <HR></HR>

      </BODY>

    Example requests/responses (**detailed** mode)::

       $ curl -X GET -i -H "Accept: application/json" "http://0.0.0.0:8775/healthcheck"
       HTTP/1.0 200 OK
       Date: Wed, 24 Aug 2016 06:11:59 GMT
       Content-Type: application/json
       Content-Length: 3480

       {
           "detailed": true,
           "gc": {
               "counts": [
                   293,
                   10,
                   5
               ],
               "threshold": [
                   700,
                   10,
                   10
               ]
           },
           "greenthreads": [
              ...
           ],
           "now": "2016-08-24 06:11:59.419267",
           "platform": "Linux-4.2.0-27-generic-x86_64-with-Ubuntu-14.04-trusty",
           "python_version": "2.7.6 (default, Jun 22 2015, 17:58:13) \\n[GCC 4.8.2]",
           "reasons": [
               {
                   "class": "HealthcheckResult",
                   "details": "Path '/tmp/dead' was not found",
                   "reason": "OK"
               }
           ],
           "threads": [
               ...
           ]
       }

       $ curl -X GET -i -H "Accept: text/html" "http://0.0.0.0:8775/healthcheck"
       HTTP/1.0 200 OK
       Date: Wed, 24 Aug 2016 06:36:07 GMT
       Content-Type: text/html; charset=UTF-8
       Content-Length: 6838

       <HTML>
       <HEAD><TITLE>Healthcheck Status</TITLE></HEAD>
       <BODY>
       <H1>Server status</H1>
       <B>Server hostname:</B><PRE>...</PRE>
       <B>Current time:</B><PRE>2016-08-24 06:36:07.302559</PRE>
       <B>Python version:</B><PRE>2.7.6 (default, Jun 22 2015, 17:58:13)
       [GCC 4.8.2]</PRE>
       <B>Platform:</B><PRE>Linux-4.2.0-27-generic-x86_64-with-Ubuntu-14.04-trusty</PRE>
       <HR></HR>
       <H2>Garbage collector:</H2>
       <B>Counts:</B><PRE>(77, 1, 6)</PRE>
       <B>Thresholds:</B><PRE>(700, 10, 10)</PRE>

       <HR></HR>
       <H2>Result of 1 checks:</H2>
       <TABLE bgcolor="#ffffff" border="1">
       <TBODY>
       <TR>
       <TH>
       Kind
       </TH>
       <TH>
       Reason
       </TH>
       <TH>
       Details
       </TH>

       </TR>
       <TR>
       <TD>HealthcheckResult</TD>
           <TD>OK</TD>
       <TD>Path &#39;/tmp/dead&#39; was not found</TD>
       </TR>
       </TBODY>
       </TABLE>
       <HR></HR>
       <H2>1 greenthread(s) active:</H2>
       <TABLE bgcolor="#ffffff" border="1">
       <TBODY>
       <TR>
           <TD><PRE>  File &#34;oslo_middleware/healthcheck/__main__.py&#34;, line 94, in &lt;module&gt;
           main()
         File &#34;oslo_middleware/healthcheck/__main__.py&#34;, line 90, in main
           server.serve_forever()
         ...
       </PRE></TD>
       </TR>
       </TBODY>
       </TABLE>
       <HR></HR>
       <H2>1 thread(s) active:</H2>
       <TABLE bgcolor="#ffffff" border="1">
       <TBODY>
       <TR>
           <TD><PRE>  File &#34;oslo_middleware/healthcheck/__main__.py&#34;, line 94, in &lt;module&gt;
           main()
         File &#34;oslo_middleware/healthcheck/__main__.py&#34;, line 90, in main
           server.serve_forever()
         ....
       </TR>
       </TBODY>
       </TABLE>
       </BODY>
       </HTML>

    Example of paste configuration:

    .. code-block:: ini

        [app:healthcheck]
        use = egg:oslo.middleware:healthcheck
        backends = disable_by_file
        disable_by_file_path = /var/run/nova/healthcheck_disable

        [pipeline:public_api]
        pipeline = healthcheck sizelimit [...] public_service

    Multiple app sections can be defined if it desired to have
    pipelines with different healthcheck configuration, example:

    .. code-block:: ini

        [composite:public_api]
        use = egg:Paste#urlmap
        / = public_api_pipeline
        /healthcheck = healthcheck_public

        [composite:admin_api]
        use = egg:Paste#urlmap
        / = admin_api_pipeline
        /healthcheck = healthcheck_admin

        [pipeline:public_api_pipeline]
        pipeline = sizelimit [...] public_service

        [pipeline:admin_api_pipeline]
        pipeline = sizelimit [...] admin_service

        [app:healthcheck_public]
        use = egg:oslo.middleware:healthcheck
        backends = disable_by_file
        disable_by_file_path = /var/run/nova/healthcheck_public_disable

        [app:healthcheck_admin]
        use = egg:oslo.middleware:healthcheck
        backends = disable_by_file
        disable_by_file_path = /var/run/nova/healthcheck_admin_disable
    """  # noqa: E501

    NAMESPACE = "oslo.middleware.healthcheck"
    HEALTHY_TO_STATUS_CODES = {
        True: webob.exc.HTTPOk.code,
        False: webob.exc.HTTPServiceUnavailable.code,
    }
    HEAD_HEALTHY_TO_STATUS_CODES = {
        True: webob.exc.HTTPNoContent.code,
        False: webob.exc.HTTPServiceUnavailable.code,
    }
    PLAIN_RESPONSE_TEMPLATE = """
{% for reason in reasons %}
{% if reason %}{{reason}}{% endif -%}
{% endfor %}
"""

    HTML_RESPONSE_TEMPLATE = """
<HTML>
<HEAD><TITLE>Healthcheck Status</TITLE></HEAD>
<BODY>
{% if detailed -%}
<H1>Server status</H1>
{% if hostname -%}
<B>Server hostname:</B><PRE>{{hostname|e}}</PRE>
{%- endif %}
<B>Current time:</B><PRE>{{now|e}}</PRE>
<B>Python version:</B><PRE>{{python_version|e}}</PRE>
<B>Platform:</B><PRE>{{platform|e}}</PRE>
<HR></HR>
<H2>Garbage collector:</H2>
<B>Counts:</B><PRE>{{gc.counts|e}}</PRE>
<B>Thresholds:</B><PRE>{{gc.threshold|e}}</PRE>
<HR></HR>
{%- endif %}
<H2>Result of {{results|length}} checks:</H2>
<TABLE bgcolor="#ffffff" border="1">
<TBODY>
<TR>
{% if detailed -%}
<TH>
Kind
</TH>
<TH>
Reason
</TH>
<TH>
Details
</TH>
{% else %}
<TH>
Reason
</TH>
{%- endif %}
</TR>
{% for result in results -%}
{% if result.reason -%}
<TR>
{% if detailed -%}
    <TD>{{result.class|e}}</TD>
{%- endif %}
    <TD>{{result.reason|e}}</TD>
{% if detailed -%}
    <TD>{{result.details|e}}</TD>
{%- endif %}
</TR>
{%- endif %}
{%- endfor %}
</TBODY>
</TABLE>
<HR></HR>
{% if detailed -%}
{% if greenthreads -%}
<H2>{{greenthreads|length}} greenthread(s) active:</H2>
<TABLE bgcolor="#ffffff" border="1">
<TBODY>
{% for stack in greenthreads -%}
<TR>
    <TD><PRE>{{stack|e}}</PRE></TD>
</TR>
{%- endfor %}
</TBODY>
</TABLE>
<HR></HR>
{%- endif %}
{% if threads -%}
<H2>{{threads|length}} thread(s) active:</H2>
<TABLE bgcolor="#ffffff" border="1">
<TBODY>
{% for stack in threads -%}
<TR>
    <TD><PRE>{{stack|e}}</PRE></TD>
</TR>
{%- endfor %}
</TBODY>
</TABLE>
{%- endif %}
{%- endif %}
</BODY>
</HTML>
"""

    def __init__(
        self,
        application: WSGIApplication | None,
        conf: dict[str, ty.Any] | cfg.ConfigOpts | None = None,
    ) -> None:
        super().__init__(application, conf)
        self.oslo_conf.register_opts(
            opts.HEALTHCHECK_OPTS, group='healthcheck'
        )
        self._show_details = self._conf_get('detailed')
        self._source_ranges = [
            ipaddress.ip_network(r)
            for r in self._conf_get('allowed_source_ranges')
        ]
        self._ignore_proxied_requests = self._conf_get(
            'ignore_proxied_requests'
        )

        # (abhishekk): Verify that if `enable_by_files` and
        # `disable_by_file` backends are not enabled at same time.
        self._verify_configured_plugins()

        if ty.TYPE_CHECKING:
            self._backends: stevedore.NamedExtensionManager[
                pluginbase.HealthcheckBaseExtension
            ]
        self._backends = stevedore.NamedExtensionManager(
            self.NAMESPACE,
            self._conf_get('backends'),
            name_order=True,
            invoke_on_load=True,
            invoke_args=(self.oslo_conf, self.conf),
        )
        self._accept_to_functor = collections.OrderedDict(
            [
                # Order here matters...
                ('text/plain', self._make_text_response),
                ('text/html', self._make_html_response),
                ('application/json', self._make_json_response),
            ]
        )
        self._accept_order = tuple(self._accept_to_functor)
        # When no accept type matches instead of returning 406 we will
        # always return text/plain (because sending an error from this
        # middleware actually can cause issues).
        self._default_accept = 'text/plain'

    def _verify_configured_plugins(self) -> None:
        backends = self._conf_get('backends')
        exclusive_plugins = ['disable_by_file', 'enable_by_files']

        if set(exclusive_plugins).issubset(set(backends)):
            raise ConfigInvalid(
                '`enable_by_files` plugin and '
                '`disable_by_file` plugin should not be '
                'enabled at the same time.'
            )

    def _conf_get(self, key: str, group: str = 'healthcheck') -> ty.Any:
        return super()._conf_get(key, group=group)

    @classmethod
    def factory(
        cls: type[base.MiddlewareType],
        global_conf: dict[str, ty.Any] | None,
        **local_conf: ty.Any,
    ) -> ty.Callable[[WSGIApplication], base.MiddlewareType]:
        raise NotImplementedError(
            'HealthcheckMiddleware should be deployed as an app, not a filter'
        )

    @classmethod
    def app_factory(
        cls,
        global_conf: dict[str, ty.Any] | None,
        **local_conf: ty.Any,
    ) -> ty_ext.Self:
        """Factory method for paste.deploy.

        :param global_conf: dict of options for all middlewares
                            (usually the [DEFAULT] section of the paste deploy
                            configuration file)
        :param local_conf: options dedicated to this middleware
                           (usually the option defined in the middleware
                           section of the paste deploy configuration file)
        """
        conf = global_conf.copy() if global_conf else {}
        conf.update(local_conf)
        middleware = cls(None, conf)
        return middleware

    @staticmethod
    def _get_threadstacks() -> list[ty.Any]:
        threadstacks = []
        try:
            active_frames = sys._current_frames()
        except AttributeError:
            pass
        else:
            buf = io.StringIO()
            for stack in active_frames.values():
                traceback.print_stack(stack, file=buf)
                threadstacks.append(buf.getvalue())
                buf.seek(0)
                buf.truncate()
        return threadstacks

    @staticmethod
    def _get_greenstacks() -> list[ty.Any]:
        greenstacks = []
        if greenlet is not None:
            buf = io.StringIO()
            for gt in _find_objects(greenlet.greenlet):
                traceback.print_stack(gt.gr_frame, file=buf)
                greenstacks.append(buf.getvalue())
                buf.seek(0)
                buf.truncate()
        return greenstacks

    @staticmethod
    def _pretty_json_dumps(contents: dict[str, ty.Any]) -> str:
        return json.dumps(contents, indent=4, sort_keys=True)

    @staticmethod
    def _are_results_healthy(
        results: list[pluginbase.HealthcheckResult],
    ) -> bool:
        for result in results:
            if not result.available:
                return False
        return True

    def _make_text_response(
        self,
        results: list[pluginbase.HealthcheckResult],
        healthy: bool,
    ) -> tuple[str, str]:
        params = {
            'reasons': [result.reason for result in results],
            'detailed': self._show_details,
        }
        body = _expand_template(self.PLAIN_RESPONSE_TEMPLATE, params)
        return (body.strip(), 'text/plain')

    def _make_json_response(
        self,
        results: list[pluginbase.HealthcheckResult],
        healthy: bool,
    ) -> tuple[str, str]:
        if self._show_details:
            body = {
                'detailed': True,
                'python_version': sys.version,
                'now': str(timeutils.utcnow()),
                'platform': platform.platform(),
                'gc': {
                    'counts': gc.get_count(),
                    'threshold': gc.get_threshold(),
                },
            }
            reasons: list[Reason] = []
            for result in results:
                reasons.append(
                    {
                        'reason': result.reason,
                        'details': result.details or '',
                        'class': reflection.get_class_name(
                            result, fully_qualified=False
                        ),
                    }
                )
            body['reasons'] = reasons
            body['greenthreads'] = self._get_greenstacks()
            body['threads'] = self._get_threadstacks()
        else:
            body = {
                'reasons': [result.reason for result in results],
                'detailed': False,
            }
        return (self._pretty_json_dumps(body), 'application/json')

    def _make_head_response(
        self,
        results: list[pluginbase.HealthcheckResult],
        healthy: bool,
    ) -> tuple[str, str]:
        return ("", "text/plain")

    def _make_html_response(
        self,
        results: list[pluginbase.HealthcheckResult],
        healthy: bool,
    ) -> tuple[str, str]:
        try:
            hostname = socket.gethostname()
        except OSError:
            hostname = None
        translated_results: list[Reason] = []
        for result in results:
            translated_results.append(
                {
                    'reason': result.reason,
                    'details': result.details or '',
                    'class': reflection.get_class_name(
                        result, fully_qualified=False
                    ),
                }
            )
        params = {
            'healthy': healthy,
            'hostname': hostname,
            'results': translated_results,
            'detailed': self._show_details,
            'now': str(timeutils.utcnow()),
            'python_version': sys.version,
            'platform': platform.platform(),
            'gc': {
                'counts': gc.get_count(),
                'threshold': gc.get_threshold(),
            },
            'threads': self._get_threadstacks(),
            'greenthreads': self._get_threadstacks(),
        }
        body = _expand_template(self.HTML_RESPONSE_TEMPLATE, params)
        return (body.strip(), 'text/html')

    @webob.dec.wsgify
    def process_request(  # type: ignore[override]
        self,
        req: webob.request.Request,
    ) -> webob.response.Response | None:
        if self._source_ranges:
            if not req.remote_addr:
                return None

            remote_addr = ipaddress.ip_address(req.remote_addr)
            for r in self._source_ranges:
                if r.version == remote_addr.version and remote_addr in r:
                    break
            else:
                # Because source ip is not included in allowed ranges, ignore
                # the request in this middleware.
                return None

        if self._ignore_proxied_requests:
            for hdr in [
                'FORWARDED',
                'FORWARDED_PROTO',
                'FORWARDED_HOST',
                'FORWARDED_FOR',
                'FORWARDED_PREFIX',
            ]:
                if req.environ.get(f"HTTP_X_{hdr}"):
                    return None

        results = [
            ext.obj.healthcheck(req.server_port)
            for ext in self._backends
            if ext.obj
        ]
        healthy = self._are_results_healthy(results)
        if req.method == "HEAD":
            functor = self._make_head_response
            status = self.HEAD_HEALTHY_TO_STATUS_CODES[healthy]
        else:
            status = self.HEALTHY_TO_STATUS_CODES[healthy]
            try:
                offers = req.accept.acceptable_offers(self._accept_order)
                accept_type = offers[0][0]
            except IndexError:
                accept_type = self._default_accept
            functor = self._accept_to_functor[accept_type]
        body, content_type = functor(results, healthy)
        return webob.response.Response(
            status=str(status),
            body=body,
            charset='UTF-8',
            content_type=content_type,
        )



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import base64
import binascii
import logging
import typing as ty

import bcrypt
import webob

from oslo_config import cfg
from oslo_middleware import base
from oslo_middleware import exceptions

if ty.TYPE_CHECKING:
    from _typeshed.wsgi import WSGIApplication
    import webob.request
    import webob.response

LOG = logging.getLogger(__name__)

OPTS = [
    cfg.StrOpt(
        'http_basic_auth_user_file',
        default='/etc/htpasswd',
        help="HTTP basic auth password file.",
    )
]

cfg.CONF.register_opts(OPTS, group='oslo_middleware')


class AuthEntry(ty.TypedDict):
    HTTP_X_USER: str
    HTTP_X_USER_NAME: str


class BasicAuthMiddleware(base.ConfigurableMiddleware):
    """Middleware which performs HTTP basic authentication on requests"""

    def __init__(
        self,
        application: WSGIApplication | None,
        conf: dict[str, ty.Any] | cfg.ConfigOpts | None = None,
    ) -> None:
        super().__init__(application, conf)
        self.auth_file = cfg.CONF.oslo_middleware.http_basic_auth_user_file
        validate_auth_file(self.auth_file)

    def format_exception(self, e: Exception) -> webob.response.Response:
        result = {'error': {'message': str(e), 'code': 401}}
        headers = [('Content-Type', 'application/json')]
        return webob.Response(
            content_type='application/json',
            status_code=401,
            json_body=result,
            headerlist=headers,
        )

    @webob.dec.wsgify
    def __call__(
        self,
        req: webob.request.Request,
    ) -> webob.response.Response | None:
        try:
            token = parse_header(req.environ)
            username, password = parse_token(token)
            req.environ.update(
                authenticate(self.auth_file, username, password)
            )
            response = req.get_response(self.application)
        except Exception as e:
            response = self.format_exception(e)

        return response


def authenticate(
    auth_file: str,
    username: str,
    password: bytes,
) -> AuthEntry:
    """Finds username and password match in Apache style user auth file

    The user auth file format is expected to comply with Apache
    documentation[1] however the bcrypt password digest is the *only*
    digest format supported.

    [1] https://httpd.apache.org/docs/current/misc/password_encryptions.html

    :param: auth_file: Path to user auth file
    :param: username: Username to authenticate
    :param: password: Password encoded as bytes
    :returns: A dictionary of WSGI environment values to append to the request
    :raises: HTTPUnauthorized, if no file entries match username/password
    """

    line_prefix = username + ':'
    try:
        with open(auth_file) as f:
            for line in f:
                entry = line.strip()
                if entry and entry.startswith(line_prefix):
                    return auth_entry(entry, password)
    except OSError as exc:
        LOG.error('Problem reading auth file: %s', exc)
        raise webob.exc.HTTPBadRequest(detail='Problem reading auth file')
    # reached end of file with no matches
    LOG.info('User %s not found', username)
    raise webob.exc.HTTPUnauthorized()


def auth_entry(entry: str, password: bytes) -> AuthEntry:
    """Compare a password with a single user auth file entry

    :param: entry: Line from auth user file to use for authentication
    :param: password: Password encoded as bytes
    :returns: A dictionary of WSGI environment values to append to the request
    :raises: HTTPUnauthorized, if the entry doesn't match supplied password or
        if the entry is crypted with a method other than bcrypt
    """
    username, crypted = parse_entry(entry)
    if not bcrypt.checkpw(password, crypted):
        LOG.info('Password for %s does not match', username)
        raise webob.exc.HTTPUnauthorized()
    return {'HTTP_X_USER': username, 'HTTP_X_USER_NAME': username}


def validate_auth_file(auth_file: str) -> None:
    """Read the auth user file and validate its correctness

    :param: auth_file: Path to user auth file
    :raises: ConfigInvalid on validation error
    """
    try:
        with open(auth_file) as f:
            for line in f:
                entry = line.strip()
                if entry and ':' in entry:
                    parse_entry(entry)
    except OSError:
        raise exceptions.ConfigInvalid(
            error_msg='Problem reading auth user file'
        )


def parse_entry(entry: str) -> tuple[str, bytes]:
    """Extrace the username and crypted password from a user auth file entry

    :param: entry: Line from auth user file to use for authentication
    :returns: a tuple of username and crypted password
    :raises: ConfigInvalid if the password is not in the supported bcrypt
    format
    """
    username, crypted_str = entry.split(':', maxsplit=1)
    crypted = crypted_str.encode('utf-8')
    if crypted[:4] not in (b'$2y$', b'$2a$', b'$2b$'):
        error_msg = (
            f'Only bcrypt digested passwords are supported for {username}'
        )
        raise webob.exc.HTTPBadRequest(detail=error_msg)
    return username, crypted


def parse_token(token: str | bytes) -> tuple[str, bytes]:
    """Parse the token portion of the Authentication header value

    :param: token: Token value from basic authorization header
    :returns: tuple of username, password
    :raises: BadRequest, if username and password could not be parsed for any
        reason
    """
    try:
        if isinstance(token, str):
            token = token.encode('utf-8')
        auth_pair = base64.b64decode(token, validate=True)
        username, password = auth_pair.split(b':', maxsplit=1)
        return (username.decode('utf-8'), password)
    except (TypeError, binascii.Error, ValueError) as exc:
        LOG.info('Could not decode authorization token: %s', exc)
        raise webob.exc.HTTPBadRequest(
            detail=('Could not decode authorization token')
        )


def parse_header(env: dict[str, str]) -> str:
    """Parse WSGI environment for Authorization header of type Basic

    :param: env: WSGI environment to get header from
    :returns: Token portion of the header value
    :raises: HTTPUnauthorized, if header is missing or if the type is not Basic
    """

    try:
        auth_header = env.pop('HTTP_AUTHORIZATION')
    except KeyError:
        LOG.info('No authorization token received')
        raise webob.exc.HTTPUnauthorized()
    try:
        auth_type, token = auth_header.strip().split(maxsplit=1)
    except (ValueError, AttributeError) as exc:
        LOG.info('Could not parse Authorization header: %s', exc)
        raise webob.exc.HTTPBadRequest(
            detail=('Could not parse Authorization header')
        )
    if auth_type.lower() != 'basic':
        error_msg = 'Unsupported authorization type "%s"'
        LOG.info(error_msg, auth_type)
        raise webob.exc.HTTPBadRequest(detail=error_msg % auth_type)
    return token



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg


HEALTHCHECK_OPTS = [
    cfg.BoolOpt(
        'detailed',
        default=False,
        help='Show more detailed information as part of the response. '
        'Security note: Enabling this option may expose '
        'sensitive details about the service being monitored. '
        'Be sure to verify that it will not violate your '
        'security policies.',
    ),
    cfg.ListOpt(
        'backends',
        default=[],
        help='Additional backends that can perform health checks and '
        'report that information back as part of a request.',
    ),
    cfg.ListOpt(
        'allowed_source_ranges',
        default=[],
        help='A list of network addresses to limit source ip allowed '
        'to access healthcheck information. Any request from ip '
        'outside of these network addresses are ignored.',
    ),
    cfg.BoolOpt(
        'ignore_proxied_requests',
        default=False,
        help='Ignore requests with proxy headers.',
    ),
]


DISABLE_BY_FILE_OPTS = [
    cfg.StrOpt(
        'disable_by_file_path',
        default=None,
        help='Check the presence of a file to determine if an '
        'application is running on a port. Used by '
        'DisableByFileHealthcheck plugin.',
    ),
]


DISABLE_BY_FILES_OPTS = [
    cfg.ListOpt(
        'disable_by_file_paths',
        default=[],
        help='Check the presence of a file based on a port to '
        'determine if an application is running on a port. '
        'Expects a "port:path" list of strings. Used by '
        'DisableByFilesPortsHealthcheck plugin.',
    ),
]


ENABLE_BY_FILES_OPTS = [
    cfg.ListOpt(
        'enable_by_file_paths',
        default=[],
        help='Check the presence of files. Used by '
        'EnableByFilesHealthcheck plugin.',
    ),
]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import argparse
from http import server
import socketserver
import typing as ty

import webob

from oslo_middleware import healthcheck

if ty.TYPE_CHECKING:
    import webob.request


class HttpHandler(server.SimpleHTTPRequestHandler):
    def do_GET(self) -> None:
        @webob.dec.wsgify
        def dummy_application(req: webob.request.Request) -> str:
            return 'test'

        app = healthcheck.Healthcheck(dummy_application, {'detailed': True})
        req = webob.Request.blank(
            "/healthcheck", accept='text/html', method='GET'
        )
        res = req.get_response(app)
        self.send_response(res.status_code)
        for header_name, header_value in res.headerlist:
            self.send_header(header_name, header_value)
        self.end_headers()
        self.wfile.write(res.body)
        self.wfile.close()


def positive_int(blob: str) -> int:
    value = int(blob)
    if value < 0:
        msg = f"{blob!r} is not a positive integer"
        raise argparse.ArgumentTypeError(msg)
    return value


def create_server(port: int = 0) -> socketserver.TCPServer:
    handler = HttpHandler
    server = socketserver.TCPServer(("", port), handler)
    return server


def main(args: ty.Sequence[str] | None = None) -> None:
    """Runs a basic http server to show healthcheck functionality."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p",
        "--port",
        help="Unused port to run the tiny"
        " http server on (or zero to select a"
        " random unused port)",
        type=positive_int,
        required=True,
    )
    parsed_args = parser.parse_args(args=args)
    server = create_server(parsed_args.port)
    print(f"Serving at port: {server.server_address[1]}")
    server.serve_forever()


if __name__ == '__main__':
    main()



# Copyright (c) 2013 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import typing as ty

from oslo_utils import uuidutils

from oslo_middleware import base

if ty.TYPE_CHECKING:
    import webob.request
    import webob.response


class CorrelationId(base.ConfigurableMiddleware):
    "Middleware that attaches a correlation id to WSGI request"

    @staticmethod
    def process_request(
        req: webob.request.Request,
    ) -> webob.response.Response | None:
        correlation_id = (
            req.headers.get("X_CORRELATION_ID") or uuidutils.generate_uuid()
        )
        req.headers['X_CORRELATION_ID'] = correlation_id
        return None
